from filternaut.exceptions import InvalidData

try:
    from rest_framework.exceptions import ParseError
    from rest_framework.filters import BaseFilterBackend
except ImportError:
    raise ImportError(
        "You must install Django REST Framework (pypi: "
        "'djangorestframework') to use Filternaut's DRF filter backend."
    )


class FilternautBackend(BaseFilterBackend):
    """
    FilternautBackend is a "custom generic filtering backend" for Django REST
    framework:
    http://www.django-rest-framework.org/api-guide/filtering/#custom-generic-filtering

    It allows straightforward filtering of a view's queryset using request
    parameters.
    """

    #: The host view must define filters at this attribute.
    filter_attr = "filternaut_filters"

    def filter_queryset(self, request, queryset, view):
        """
        Decide whether to apply the filters defined by
        ``view.filternaut_filters`` on the argued queryset. If the filters
        parse correctly, ``is_valid`` is called. If not, ``is_invalid`` is
        called
        """
        try:
            filters = getattr(view, self.filter_attr)
        except AttributeError:
            raise AttributeError(
                f"View {view} requires attribute '{self.filter_attr}' to use FilternautBackend"
            )

        if callable(filters):
            filters = filters(request)

        try:
            query = filters.parse(request.query_params)
            return self.is_valid(request, queryset, query)
        except InvalidData as ex:
            return self.is_invalid(request, queryset, ex.errors)

    def is_valid(self, request, queryset, query):
        """
        Apply Q-object ``query`` to ``queryset``. Provided for convenience when
        subclassing.
        """
        return queryset.filter(query)

    def is_invalid(self, request, queryset, errors):
        """
        Raise a ParseError containing the filter errors. This results in a 400
        Bad Request whose body details those errors. Provided for convenience
        when subclassing.
        """
        raise ParseError(errors)



import tempfile
from unittest import TestCase

import pytest
from django.forms import CharField, IntegerField
from django.utils.datastructures import MultiValueDict

import filternaut
from filternaut import Filter, Optional
from filternaut.exceptions import InvalidData
from filternaut.filters import (
    CharFilter,
    ChoiceFilter,
    ComboFilter,
    FieldFilter,
    FilePathFilter,
    IntegerFilter,
    RegexFilter,
)
from tests.util import NopeFilter, assert_parsed_ok, flatten_qobj


class FilterTests(TestCase):
    def test_kwargless_simple_instantiation(self):
        f = Filter("fieldname")
        assert f.source == "fieldname"
        assert f.dest == "fieldname"

    def test_kwarg_instantiation(self):
        f = Filter(dest="fieldname")
        assert f.source == "fieldname"
        assert f.dest == "fieldname"

        f = Filter(source="sourcename", dest="fieldname")
        assert f.source == "sourcename"
        assert f.dest == "fieldname"

    def test_getsourcevalue_does_what_it_says_on_the_tin(self):
        f = Filter("fieldname")
        bar = f.get_source_value("foo", {"foo": "bar"})
        assert bar == "bar"

    def test_getsourcevalue_raises_keyerrors(self):
        f = Filter("fieldname")
        with self.assertRaises(KeyError):
            f.get_source_value("legume", {"foo": "bar"})

    def test_required_filter_is_satisfied_by_one_key_being_present(self):
        f = Filter("fieldname", required=True, lookups=["gte"])
        data = {"fieldname__gte": "foo", "fieldname__lte": "bar"}
        assert_parsed_ok(f.parse(data))

    def test_required_filter_requires_at_least_one_key(self):
        """
        One filter can have many possible sources due to the different lookup
        affixes. One of them being present should be enough to satisfy a
        'required' field.
        """
        f = Filter("fieldname", required=True, lookups=["contains"])
        data = {"fieldname__gte": "foo", "fieldname__lte": "bar"}
        with pytest.raises(InvalidData) as exc_info:
            f.parse(data)
        assert "fieldname" in exc_info.value.errors

    def test_lookups_can_be_provided_as_a_string(self):
        f = Filter("fieldname", lookups="contains")
        assert f.lookups == ["contains"]

    def test_lookups_can_be_provided_as_a_comma_separated_string(self):
        f = Filter("fieldname", lookups="lt,lte,gt,gte")
        assert f.lookups == ["lt", "lte", "gt", "gte"]

    def test_negating_filters(self):
        f = Filter("fieldname")
        assert f.negate is False
        f = ~f
        assert f.negate is True
        f = ~f
        assert f.negate is False

    def test_one_source_to_many_dests(self):
        filters = (
            Filter(source="search", dest="username")
            | Filter(source="search", dest="email")
            | Filter(source="search", dest="name")
        )
        query = filters.parse(dict(search="mynbaev-scheiner"))
        expected = dict(
            username="mynbaev-scheiner",
            email="mynbaev-scheiner",
            name="mynbaev-scheiner",
        )
        actual = dict(flatten_qobj(query))
        assert expected == actual

    def test_many_sources_one_dest(self):
        filters = (
            Filter(source="nominee_1", dest="field")
            | Filter(source="nominee_2", dest="field")
            | Filter(source="nominee_3", dest="field")
        )
        query = filters.parse(dict(nominee_1="one", nominee_2="two", nominee_3="three"))
        expected = dict(field="three")  # last one wins
        actual = dict(flatten_qobj(query))
        assert expected == actual

    def test_multivaluedict_as_source(self):
        """
        We don't expect multiple values to be pulled from a multi-value source
        without explicit instruction
        """
        filters = Filter("name")
        data = MultiValueDict(dict(name=["foo", "bar"]))
        query = filters.parse(data)
        expected = dict(name="bar")  # last one wins
        actual = dict(flatten_qobj(query))
        assert expected == actual

    def test_multivaluedict_as_source_when_many_values_required(self):
        """
        Having a dest ending in '__in' /does/ constitute explicit instruction
        to pull multiple values from a source.
        """
        filters = Filter("field", lookups=["in"])
        data = MultiValueDict(dict(field=["foo", "bar"]))
        query = filters.parse(data)
        expected = dict(field__in=["foo", "bar"])
        actual = dict(flatten_qobj(query))
        assert expected == actual

    def test_multivaluedict_not_used_for_nonlisty_filtering(self):
        filters = Filter("field", lookups=["exact", "in"])

        data = MultiValueDict(dict(field=["foo", "bar"]))
        query = filters.parse(data)
        expected = dict(field="bar")  # last one wins
        actual = dict(flatten_qobj(query))
        assert expected == actual

        data = MultiValueDict(dict(field__in=["foo", "bar"]))
        query = filters.parse(data)
        expected = dict(field__in=["foo", "bar"])
        actual = dict(flatten_qobj(query))
        assert expected == actual


class TestSourceDestPairCreation(TestCase):
    def test_no_lookup_same_source_and_dest(self):
        filter = Filter("field", lookups=[None])
        expected = [
            ("field", "field"),
        ]
        actual = list(filter.source_dest_pairs())
        assert expected == actual

    def test_gte_lookup_same_source_and_dest(self):
        filter = Filter("field", lookups=["gte"])
        expected = [("field__gte", "field__gte"), ("field", "field__gte")]
        actual = list(filter.source_dest_pairs())
        assert sorted(expected) == sorted(actual)

    def test_no_lookup_differing_source_and_dest(self):
        filter = Filter(source="something", dest="field", lookups=[None])
        expected = [
            ("something", "field"),
        ]
        actual = list(filter.source_dest_pairs())
        assert expected == actual

    def test_lte_lookup_differing_source_and_dest(self):
        filter = Filter(source="something", dest="field", lookups=["lte"])
        expected = [("something__lte", "field__lte"), ("something", "field__lte")]
        actual = list(filter.source_dest_pairs())
        assert sorted(expected) == sorted(actual)


class ParsingTests(TestCase):
    def test_parse_simple_source(self):
        filter = Filter("word")
        data = {"word": "a_value", "unrelated_key": ""}
        query = filter.parse(data)
        flat = dict(flatten_qobj(query))

        assert "word" in flat
        assert flat["word"] == "a_value"

    def test_parse_source_with_lookups(self):
        filter = Filter("word", lookups=["gte", "gt", "lte", "lt"])
        data = {"word__gte": 1, "word__gt": 2, "word__lte": 3, "word__lt": 4}

        query = filter.parse(data)
        flat = dict(flatten_qobj(query))

        for key in data:
            assert key in flat
        for key, value in data.items():
            assert flat[key] == value

    def test_more_extensive_parsing(self):
        filters = (
            Filter("field_1", lookups=["contains", "icontains"])
            | Filter(source="field_2", dest="field_2a")
            | Filter(source="field_3", dest="f3", lookups=["lt", "gt"])
            | Filter(source="id", dest="relationship__spanning__id")
        )
        data = {
            "field_1__contains": "first",
            "field_2": "second",
            "field_3__lt": "third",
            "field_3__gt": "fourth",
            "id": "fifth",
        }
        expected = {
            "field_1__contains": "first",
            "field_2a": "second",
            "f3__lt": "third",
            "f3__gt": "fourth",
            "relationship__spanning__id": "fifth",
        }
        query = filters.parse(data)
        actual = dict(flatten_qobj(query))
        assert expected == actual


class FieldFilterTests(TestCase):
    def test_choicefilter_choices(self):
        choices = (("alai", "Alai"), ("petra", "Petra"))
        filter = ChoiceFilter(choices=choices, dest="name")

        with pytest.raises(InvalidData):
            filter.parse({"name": "bean"})

        # no exception
        filter.parse({"name": "alai"})

    def test_choicefilter_instantiation_styles(self):
        fieldname = "fieldname"
        choices = ["choice1", "choice2"]

        filters = [
            # no kwargs
            ChoiceFilter(fieldname, choices),
            # arg fieldname, kwarg choices
            ChoiceFilter(fieldname, choices=choices),
            # all kwargs
            ChoiceFilter(source=fieldname, dest=fieldname, choices=choices),
        ]

        for filter in filters:
            assert filter.source == fieldname
            assert filter.field.choices == choices

    def test_regexfilter(self):
        filter = RegexFilter("fieldname", r"\d+")
        with pytest.raises(InvalidData) as exc_info:
            filter.parse({"fieldname": "alpha"})
        assert "fieldname" in exc_info.value.errors

        filter = RegexFilter("fieldname", r"\d+")
        query = filter.parse({"fieldname": "1008346"})
        assert_parsed_ok(query)

    def test_filepathfilter(self):
        # this isn't a great test, since we just instantiate the filter.
        # however, proper testing requires monkeying about changing the test
        # system's filesystem; and, being a subclass of choicefilter, that
        # class' tests should cover us.
        path = tempfile.gettempdir()
        FilePathFilter("fieldname", path)

    def test_combofilter(self):
        filter = ComboFilter(
            "fieldname",
            fields=[
                CharField(min_length=2, max_length=4),
                CharField(min_length=3, max_length=6),
            ],
        )
        invalids = "", "a", "aa", "aaaaa", "aaaaaa", "aaaaaaa"
        valids = "aaa", "aaaa"

        for invalid in invalids:
            with pytest.raises(InvalidData) as exc_info:
                filter.parse({"fieldname": invalid})
            assert "fieldname" in exc_info.value.errors

        for valid in valids:
            query = filter.parse({"fieldname": valid})
            assert_parsed_ok(query)

    def test_multivaluefilter(self):
        # TODO this field is relatively complex. have not written a test for it
        # yet.
        pass

    def test_instantiate_all_filterfields_without_special_args(self):
        # ChoiceFilter, FilePathFilter, and some others have their own specific
        # tests, because their fields' constructors, and therefore their own
        # constructors, require additional arguments
        filterfields = (
            "BooleanFilter",
            "CharFilter",
            "DateFilter",
            "DateTimeFilter",
            "DecimalFilter",
            "EmailFilter",
            "FloatFilter",
            "ImageFilter",
            "IntegerFilter",
            "MultipleChoiceFilter",
            "NullBooleanFilter",
            "SlugFilter",
            "SplitDateTimeFilter",
            "TimeFilter",
            "TypedChoiceFilter",
            "URLFilter",
            "GenericIPAddressFilter",
            "TypedMultipleChoiceFilter",
        )

        for f in filterfields:
            klass = getattr(filternaut.filters, f)
            klass("fieldname")

    def test_listlike_values_cleaned_individually(self):
        f = CharFilter("fieldname", lookups="exact,in")
        data = MultiValueDict(
            {
                "fieldname": ["single value"],  # MVDict treats 1-list as single
                "fieldname__in": ["multiple", "values"],
            }
        )
        expected = {
            "fieldname": "single value",
            "fieldname__in": ["multiple", "values"],
        }
        query = f.parse(data)
        actual = dict(flatten_qobj(query))
        assert expected == actual


class DefaultValueTests(TestCase):
    def test_default_value_used_if_no_sourcedata_found(self):
        filters = Filter("count", lookups=["lte", "gte"], default=3)
        query = filters.parse({})  # no value for 'count'

        expected = {"count__exact": 3}
        actual = dict(flatten_qobj(query))

        assert expected == actual

    def test_default_lookup_is_exact(self):
        f = Filter("count", default=3)
        assert f.default_lookup == "exact"

    def test_other_lookups_ignored_when_default_used(self):
        filters = Filter("count", lookups=["lte", "gte"], default=3)
        query = filters.parse({})  # no value for 'count'

        keys = dict(flatten_qobj(query)).keys()

        assert "lte" not in keys
        assert "gte" not in keys

    def test_default_lookup_type_can_be_changed(self):
        filters = Filter(
            "count", lookups=["lte", "gte"], default=3, default_lookup="gt"
        )
        query = filters.parse({})  # no value for 'count'

        expected = {"count__gt": 3}
        actual = dict(flatten_qobj(query))

        assert expected == actual

    def test_default_ignored_if_sourcedata_found(self):
        filters = Filter(
            "count", lookups=["lte", "gte"], default=3, default_lookup="foobarbaz"
        )
        query = filters.parse({"count__gte": 4})

        expected = {"count__gte": 4}
        actual = dict(flatten_qobj(query))

        assert expected == actual

    def test_default_can_be_callable(self):
        filters = Filter("count", lookups=["lte", "gte"], default=lambda: 3)
        query = filters.parse({})  # no value for 'count'

        expected = {"count__exact": 3}
        actual = dict(flatten_qobj(query))

        assert expected == actual


class OptionalTests(TestCase):
    def setUp(self):
        self.filters = Optional(
            Filter("one", required=True), Filter("two"), Filter("three", required=True)
        )
        self.unrelated = Filter("ten") & (Filter("eleven") | Filter("twelve"))

    def test_no_values_present(self):
        data = dict()  # no data
        query = self.filters.parse(data)
        assert_parsed_ok(query)

    def test_all_values_present(self):
        data = dict(one=1, two=2, three=3)
        query = self.filters.parse(data)
        assert_parsed_ok(query)

    def test_only_required_values_present(self):
        data = dict(one=1, three=3)
        query = self.filters.parse(data)
        assert_parsed_ok(query)

    def test_only_nonrequired_values_present(self):
        data = dict(two=2)
        with pytest.raises(InvalidData) as exc_info:
            self.filters.parse(data)
        assert "one" in exc_info.value.errors
        assert "three" in exc_info.value.errors
        assert "__all__" in exc_info.value.errors

    def test_one_required_filter_missing(self):
        data = dict(two=2, three=3)
        with pytest.raises(InvalidData) as exc_info:
            self.filters.parse(data)
        assert "one" in exc_info.value.errors
        assert "three" not in exc_info.value.errors
        assert "__all__" in exc_info.value.errors

    def test_ANDed_with_unrelated(self):
        filters = self.filters & self.unrelated

        data = dict()  # no data
        query = filters.parse(data)
        assert_parsed_ok(query)

        data = dict(one=1, three=3)  # all required data
        query = filters.parse(data)
        assert_parsed_ok(query)

        data = dict(ten=10)  # only some unrelated
        query = filters.parse(data)
        assert_parsed_ok(query)

        data = dict(ten=10, eleven=11, twelve=12)  # only all unrelated
        query = filters.parse(data)
        assert_parsed_ok(query)

    def test_validation_errors_are_not_silenced(self):
        filters = Optional(
            NopeFilter("a", required=True),
            NopeFilter("b", required=True),
            NopeFilter("c"),
        )
        data = dict(a=1, b=2, c=3)
        with pytest.raises(InvalidData) as exc_info:
            filters.parse(data)
        assert "a" in exc_info.value.errors
        assert "b" in exc_info.value.errors
        assert "c" in exc_info.value.errors

        data = dict()
        assert_parsed_ok(filters.parse(data))


def boolean_tests():
    filter = filternaut.filters.BooleanFilter("approved")

    valid_values = "1", "0", "true", "false", "True", "False", True, False
    for value in valid_values:
        query = filter.parse({"approved": value})
        assert_parsed_ok(query)


class MultiValueWithNoneTests(TestCase):
    def get_output_of_field(self, isnull=False, **data):
        # previously you had to pass a required=False field instance into FieldFilter.
        old_style = FieldFilter(
            "rank",
            lookups="in",
            none_to_isnull=isnull,
            field=IntegerField(required=False),
        )
        # now you can work with the field variants directly.
        new_style = IntegerFilter("rank", lookups="in", none_to_isnull=isnull)

        data = MultiValueDict(data)
        old_query = old_style.parse(data)
        new_query = new_style.parse(data)

        # check both styles produce the same results
        assert old_query == new_query

        # specific tests make their own assertions with this return value too.
        return dict(new_query.children)

    def test_enabled_with_regular_value(self):
        actual = self.get_output_of_field(rank=["1", "2", "3"], isnull=True)
        expected = {"rank__in": [1, 2, 3]}
        assert actual == expected

    def test_enabled_with_none_value_and_regulars(self):
        actual = self.get_output_of_field(rank=["1", "2", "3", ""], isnull=True)
        expected = {"rank__in": [1, 2, 3], "rank__isnull": True}
        assert actual == expected

    def test_enabled_with_only_none_value(self):
        actual = self.get_output_of_field(rank=[""], isnull=True)
        expected = {"rank__isnull": True}
        assert actual == expected

    def test_disabled_with_regular_value(self):
        actual = self.get_output_of_field(rank=["1", "2", "3"], isnull=False)
        expected = {"rank__in": [1, 2, 3]}
        assert actual == expected

    def test_disabled_with_none_value_and_regulars(self):
        with pytest.raises(InvalidData) as exc_info:
            self.get_output_of_field(rank=["1", "2", "3", ""], isnull=False)
        assert "rank" in exc_info.value.errors
        assert exc_info.value.errors["rank"] == [
            "Validation error at 4th item: This field is required."
        ]

    def test_disabled_with_only_none_value(self):
        with pytest.raises(InvalidData) as exc_info:
            self.get_output_of_field(rank=[""], isnull=False)
        assert "rank" in exc_info.value.errors
        assert exc_info.value.errors["rank"] == [
            "Validation error at 1st item: This field is required."
        ]

    def test_ben(self):
        source = MultiValueDict({"parent_id__in": [1, 2, None]})
        filters = filternaut.filters.IntegerFilter(
            "parent_id", lookups="in", none_to_isnull=True
        )
        parsed = filters.parse_to_dict(source)
        pass



from unittest import mock

from django.contrib.auth.models import User

from filternaut.filters import EmailFilter

try:
    from rest_framework import generics
    from rest_framework.test import APIRequestFactory, APITestCase

    from filternaut.drf import FilternautBackend
except ImportError:
    import pytest

    pytest.mark.skip(
        "Django REST Framework must be installed to test DRF integration",
        allow_module_level=True,
    )

    # without an APITestCase we get a NameError
    class APITestCase:
        pass


class IntegrationTests(APITestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.request = self.factory.get("/users/", data={"email": "user@example.org"})

    def test_backend_requires_filterattr(self):
        class UserView(generics.ListAPIView):
            queryset = User.objects.all()
            filter_backends = (FilternautBackend,)

        queryset = mock.Mock()
        uv = UserView()
        self.request = uv.initialize_request(self.request)
        uv.request = self.request
        with self.assertRaises(AttributeError):
            uv.filter_queryset(queryset)

    def test_backend_filters_queryset(self):
        class UserView(generics.ListAPIView):
            queryset = User.objects.all()
            filter_backends = (FilternautBackend,)
            filternaut_filters = EmailFilter("email")

        queryset = mock.Mock()
        uv = UserView()
        self.request = uv.initialize_request(self.request)
        uv.request = self.request
        uv.filter_queryset(queryset)
        assert queryset.filter.called

    def test_callable_filterattr(self):
        class UserView(generics.ListAPIView):
            queryset = User.objects.all()
            filter_backends = (FilternautBackend,)

            def filternaut_filters(self, request):
                return EmailFilter("email")

        queryset = mock.Mock()
        uv = UserView()
        self.request = uv.initialize_request(self.request)
        uv.request = self.request
        uv.filter_queryset(queryset)
        assert queryset.filter.called



from collections import defaultdict
from pprint import pformat


class FilternautException(Exception):
    pass


class InvalidData(FilternautException):
    errors = None

    def __init__(self, errors):
        if isinstance(errors, defaultdict):
            errors = dict(errors)
        super().__init__(pformat(errors))
        self.errors = errors



from collections import defaultdict, namedtuple
from copy import deepcopy
from functools import reduce
from operator import and_, or_

from django.core.exceptions import ValidationError
from django.db.models import Q

from filternaut.exceptions import InvalidData
from filternaut.tree import Leaf, Tree
from filternaut.util import is_listlike


class FilterTree(Tree):
    """
    FilterTrees instances are the result of ORing or ANDing Filter instances,
    or other FilterTree instances, together.

    FilterTree provides a simple API for dealing with a set of filters. They
    may also be iterated over to gain access to the Filter instances directly.
    """

    def __init__(self, negate=False, *args, **kwargs):
        self.negate = negate
        super().__init__(*args, **kwargs)

    def __invert__(self):
        inverted = deepcopy(self)
        inverted.negate = not self.negate
        return inverted

    def __repr__(self):
        children = ", ".join(f.source for f in self)
        return f"<{type(self).__name__} {children}>"

    def parse(self, data):
        """
        Have this tree of filters convert ``data`` into a Django Q-object.
        """
        errors = {}
        # do this with two try/except so we can collect errors from both
        # branches
        try:
            left_q = self.left.parse(data)
        except InvalidData as ex:
            errors.update(ex.errors)

        try:
            right_q = self.right.parse(data)
        except InvalidData as ex:
            errors.update(ex.errors)

        if errors:
            raise InvalidData(errors=errors)

        return self.operator(left_q, right_q)


class Constraint(FilterTree):
    """
    Base class for making collective rules over several filters.

    For example, at least two of the filters must be used, or only one of the
    filters may be used, etc.

    This class book-keeps which filters were used as it parses the incoming
    data. Subclasses can use this to decide whether to add or remove errors.
    """

    tree_class = FilterTree

    #: Subclasses will receive a ``FilterUse`` for each child filter. This
    #: should be used to adjust validation as necessary. See
    #: :py:meth:`Constraint.apply_constraint`.
    FilterUse = namedtuple("FilterUse", "filter,valid,missing")

    def __init__(self, left, *rest):
        operator = and_  # OR does not make sense with group-require
        if not rest:
            # TODO if left is a tree, walk it instead of complaining
            raise ValueError("Optional has no effect on a single filter")
        right = reduce(and_, rest)
        super().__init__(False, operator, left, right)

    def parse(self, data):
        """
        Return Q-object for ``data``
        """
        filters = list(self)
        errors = defaultdict(list)
        query = Q()
        report = []

        # this is the same as the regular parse(), but with book-keeping to
        # populate `report`.
        for filter in filters:
            try:
                this_query = filter.parse(data)
                if this_query:
                    report.append(self.FilterUse(filter, valid=True, missing=False))
                else:
                    report.append(self.FilterUse(filter, valid=None, missing=True))
                query = self.operator(query, this_query)

            except InvalidData as ex:
                relevant_errors = ex.errors.get(filter.source, {})
                # TODO find a less terrible way of determining this
                is_missing = "This field is required" in relevant_errors
                report.append(self.FilterUse(filter, valid=False, missing=is_missing))
                errors.update(ex.errors)

        self.apply_constraint(report, errors)
        if errors:
            raise InvalidData(errors)

        return query

    def apply_constraint(self, report, errors):
        """
        Subclasses should examine how filters were used in ``report`` and
        mutate ``errors`` as necessary.
        """
        raise NotImplementedError()


class Optional(Constraint):
    """
    Filters included underneath Optional have their required=True configuration
    ignored as long as all those filters are missing. If some but not all are
    present, then required=True is observed, and those filters that are missing
    become invalid. If all filters under Optional have valid source data, they
    are valid as a whole.

    This is useful for situations where you want to require one field if
    another is present. For example, requiring ``last_name`` if ``first_name``
    is present, but also allowing neither. In this case, you would mark both
    with required=True, and wrap them in Optional::

        Optional(
            Filter('first_name', required=True),
            Filter('last_name', required=True),
        )

    Generally, filters underneath Optional will have required=True, however it
    isn't necessary; consider adding 'middle name' to the above example.
    """

    def apply_constraint(self, report, errors):
        valid = [r.filter for r in report if r.valid]
        triggers = [r.filter for r in report if r.filter.required and r.missing]

        # some filters have values, but not all required filters have values.
        if any(valid) and any(triggers):
            sources = ", ".join(sorted(r.filter.source for r in report))
            errors["__all__"].append(
                f"If any of {sources} are provided, all must be provided"
            )

        else:
            for filter in triggers:
                del errors[filter.source]


class OneOf(Constraint):
    """
    Only one of the child filters can be specified at a time. Specifying two or
    more causes a validation error. Use like so::

        filters = OneOf(
            Filter("foo"),
            Filter("bar"),
            Filter("baz"),
        )
    """

    def apply_constraint(self, report, errors):
        sources = ", ".join(sorted(r.filter.source for r in report))
        valid_count = sum(r.valid for r in report)
        extra_error = None

        if valid_count == 0:
            errors["__all__"].append(f"One of {sources} must be provided")
        elif valid_count > 1:
            errors["__all__"].append(f"Only one of {sources} can be provided")
        elif valid_count == 1:
            for r in report:
                if r.filter.required and r.missing:
                    del errors[r.filter.source]


class Filter(Leaf):
    """
    A Filter instance builds a django.db.models.Q object by pulling a value
    from arbitrary native data, e.g. a set of query params.

    It can be ORed or ANDed together with other Filter instances in the same
    way Q objects can.
    """

    #: Filters combine into FilterTree instances
    tree_class = FilterTree

    def __init__(self, dest, **kwargs):
        self.dest = dest
        self.source = kwargs.get("source", dest)
        self.lookups = kwargs.get("lookups", ["exact"])
        self.multivalue_lookups = kwargs.get("multivalue_lookups", ["in"])
        self.required = kwargs.get("required", False)
        self.negate = kwargs.get("negate", False)
        self.none_to_isnull = kwargs.get("none_to_isnull", False)

        # None is a valid default -- consider exclude(groups=None) -- so use
        # the absence or presence of self.default to indicate whether a default
        # should be used.
        if "default" in kwargs:
            self.default = kwargs["default"]
            self.default_lookup = kwargs.get("default_lookup", "exact")

        # accept lookups as a comma-separated string.
        if isinstance(self.lookups, str):
            self.lookups = self.lookups.split(",")

        if self.none_to_isnull and self.lookups != ["in"]:
            raise TypeError("When none_to_isnull=True, you must also use lookups='in'")

    def __invert__(self):
        """
        Invert the sense of this filter.
        """
        inverted = self.copy()
        inverted.negate = not self.negate
        return inverted

    def __repr__(self):
        cls = type(self)
        desc = self.source
        if self.lookups != ["exact"]:
            desc = desc + "__" + "/".join(self.lookups)
        return f"<{cls.__name__} {desc}>"

    def parse_to_dict(self, data):
        """
        Look through the provided dict-like data for keys which match this
        Filter's source. This includes keys containing lookup affixes such as
        'contains' or 'lte'::

            {"created_date__gte": "2020-01-01..."}

        If you just want to plug Filternaut into Django's ORM, call
        :py:meth:`Filter.parse` instead.

        If you can't get Filternaut to do what you want, you might have an
        easier time post-processing the output of ``parse_to_dict`` and
        dropping it directly into the ORM's ``filter()`` yourself. Note that
        this method is only available on individual filters and not on a
        combination (such as ``Filter("a") & Filter("b")``), so this approach
        only takes you so far.
        """
        source_pairs = self.source_value_pairs(data)
        dest_pairs, errors = self.dest_value_pairs(source_pairs)

        # handle default value
        if not source_pairs and hasattr(self, "default"):
            dest_pairs = (self.default_dest_value_pair(),)

        # if required, check if satisfied
        if not source_pairs and self.required:
            if self.source not in errors:
                errors[self.source] = []
            errors[self.source].append("This field is required")

        if errors:
            raise InvalidData(errors=errors)

        return dict(dest_pairs)

    def parse(self, data):
        """
        Return Q-object which can be used with Django ORM.

        In the general case this is just Q() wrapped around the output of
        :py:meth:`Filter.parse_to_dict`; they are different representations of the same
        information::

            parse_to_dict() -> {"id": 1}
            parse()         -> Q(id=1)

        When none-to-isnull conversion is enabled and relevant, this method
        returns a query which differs slightly from the dict representation::

            parse_to_dict() -> {"id__in": [1, 2, None]}
            parse()         -> Q(id__in=[1, 2]) | Q(id__isnull=True)

        """
        filter_dict = self.parse_to_dict(data)

        trigger = f"{self.dest}__in"
        do_isnull_conversion = (
            self.none_to_isnull
            and trigger in filter_dict
            and None in filter_dict[trigger]
        )
        if not do_isnull_conversion:
            # normal case, return Q directly from filter-dict
            query = Q(**filter_dict)
            return query

        # when none-to-isnull conversion is active, remove None from __in and
        # replace by ORing with __isnull=True.
        filter_dict[trigger] = [v for v in filter_dict[trigger] if v is not None]
        if not filter_dict[trigger]:
            filter_dict.pop(trigger)  # vals were all Nones; ditch key entirely
        isnull_dest = f"{self.dest}__isnull"
        return Q(**filter_dict) | Q(**{f"{self.dest}__isnull": True})

    def clean(self, value):
        """
        Validate and normalise ``value`` for use in filtering. This
        implementation is a no-op; subclasses may do more work here.
        """
        return value

    def dest_is_many(self, dest):
        return any(dest.endswith(f"__{x}") for x in self.multivalue_lookups)

    def source_dest_pairs(self):
        """
        For each lookup in self.lookups, such as 'contains' or 'lte', combine
        it with this field's source and dest, returning e.g.
        (username__contains, account_name__contains)

        If any lookup is None, that pair becomes (source, dest)

        If there is only one lookup, two pairs are listed containing the source
        both with and without the lookup. This allows source data to omit the
        lookup from the key, e.g. providing 'email' to the filter
        Filter('email', lookups=['iexact']).
        """
        pairs = []
        for lookup in self.lookups:
            if lookup in (None, "exact"):
                source = self.source
                dest = self.dest
            else:
                source = f"{self.source}__{lookup}"
                dest = f"{self.dest}__{lookup}"
            pairs.append((source, dest))

        # allow source data to omit the lookup if only one lookup listed.
        if len(pairs) == 1:
            lookup = self.lookups[0]
            if lookup not in (None, "exact"):
                dest = f"{self.dest}__{self.lookups[0]}"
                pairs.append((self.source, dest))

        return pairs

    def source_value_pairs(self, data):
        """
        Return a list of two-tuples containing valid sources for this filter --
        made by combining this filter's source and the various lookups -- and
        their values, as pulled from the data handed to parse.

        Sources with no found value are excluded.
        """
        pairs = []
        for source, dest in self.source_dest_pairs():
            try:
                many = self.dest_is_many(dest)
                value = self.get_source_value(source, data, many)
                pairs.append((source, value))
            except KeyError:
                pass
        return pairs

    def dest_value_pairs(self, sourcevalue_pairs):
        """
        Return two values:
            - A list of two-tuples containing dests (ORM relation/field names)
              and their values.
            - A dictionary of errors, keyed by the source which they originated
              from.
        """
        sourcedest_map = dict(self.source_dest_pairs())
        pairs = []
        errors = {}
        for source, value in sourcevalue_pairs:
            try:
                value = self.clean(value)
            except ValidationError as ex:
                errors[source] = ex.messages
                continue

            dest = sourcedest_map[source]
            pairs.append((dest, value))

        return pairs, errors

    def default_dest_value_pair(self):
        """
        Construct a default dest/value pair to be used if no source data was
        found during parsing (and if this filter has default=True).
        """
        dest = f"{self.dest}__{self.default_lookup}"
        default = self.default() if callable(self.default) else self.default
        return (dest, default)

    def get_source_value(self, key, data, many=False):
        """
        Pull ``key`` from ``data``.

        When ``many`` is True, a list of values is returned. Otherwise, a
        single value is returned.
        """
        if many is False:
            return data[key]
        elif hasattr(data, "getlist"):  # Django querydict, multivaluedict
            if key not in data:
                raise KeyError(repr(key))
            return data.getlist(key)
        else:
            # only a single value, but many=True, so return as list.
            return [data[key]]



from functools import reduce
from operator import and_, or_
from unittest import TestCase

import pytest

from filternaut import Filter, FilterTree, Optional
from filternaut.exceptions import InvalidData
from filternaut.tree import Leaf, Tree
from tests.util import NopeFilter, assert_parsed_ok


class OperatorTests(TestCase):
    def test_leaves_OR_into_trees(self):
        tree = Leaf() | Leaf()
        assert isinstance(tree, Tree)

    def test_leaves_AND_into_trees(self):
        tree = Leaf() & Leaf()
        assert isinstance(tree, Tree)

    def test_many_leaves_combine_into_tree(self):
        leaves = [Leaf() for n in range(100)]

        def operators():
            while True:
                yield or_
                yield and_

        # alternate between ORing and ANDing leaves together
        tree = reduce(lambda l, r: next(operators())(l, r), leaves)

        expected = leaves
        actual = [l for l in tree]
        assert isinstance(tree, Tree)
        assert len(expected) == len(actual)


class FilterTreeTests(TestCase):
    """
    Subclasses of Leaf and Tree should exhibit treelike behaviour.
    """

    def test_filters_OR_into_filtersets(self):
        filters = Filter("afield") | Filter("anotherfield")
        assert isinstance(filters, FilterTree)

    def test_filters_AND_into_filtersets(self):
        filters = Filter("afield") & Filter("anotherfield")
        assert isinstance(filters, FilterTree)

    def test_errors_from_filtertree(self):
        """
        FilterTree's errors should be the errors of the filters it contains.
        """
        filters = NopeFilter("one") | NopeFilter("two") | NopeFilter("three")
        with pytest.raises(InvalidData) as exc_info:
            filters.parse(dict(one=1, two=2, three=3))
        assert "one" in exc_info.value.errors
        assert "two" in exc_info.value.errors
        assert "three" in exc_info.value.errors

    def test_validity_from_filtertree(self):
        """
        FilterTree's validity should be a proxy to the validities of the
        filters it contains.
        """
        filters = Filter("one") | Filter("two") | Filter("three")
        query = filters.parse(dict())
        assert_parsed_ok(query)

    def test_Optional_combines_well(self):
        """
        Combine Optional with Filters and FilterTrees in various combinations.
        """
        f1, f2, f3, f4, f5 = map(Filter, ["1", "2", "3", "4", "5"])
        o1 = Optional(f1, f2)
        o2 = Optional(f3, f4)

        left_or = o1 | f3
        left_and = o1 & f3
        right_or = f3 | o1
        right_and = f3 & o1
        both_or = o1 | o2
        both_and = o1 & o2

        filters_list = (left_or, left_and, right_or, right_and, both_or, both_and)
        filters_with_123 = left_or, left_and, right_or, right_and
        filters_with_1234 = both_or, both_and
        for filters in filters_list:
            assert filters.__class__ == FilterTree

        for filters in filters_with_123:
            required = f1, f2, f3
            assert set(required) == set(filters)

        for filters in filters_with_1234:
            required = f1, f2, f3, f4
            assert set(required) == set(filters)



import json

from django.contrib.auth.models import User
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest
from django.test import Client, TestCase

from filternaut.exceptions import InvalidData
from filternaut.filters import CharFilter, ChoiceFilter, EmailFilter

try:
    from django.test import RequestFactory
except ImportError:

    class RequestFactory(Client):
        """
        Django 1.2 does not have RequestFactory. credit:
        https://djangosnippets.org/snippets/963/
        """

        def request(self, **request):
            """
            Similar to parent class, but returns the request object as soon as
            it has created it.
            """
            environ = {
                "HTTP_COOKIE": self.cookies,
                "PATH_INFO": "/",
                "QUERY_STRING": "",
                "REQUEST_METHOD": "GET",
                "SCRIPT_NAME": "",
                "SERVER_NAME": "testserver",
                "SERVER_PORT": 80,
                "SERVER_PROTOCOL": "HTTP/1.1",
            }
            environ.update(self.defaults)
            environ.update(request)
            return WSGIRequest(environ)


def user_to_native(user):
    return dict(
        id=user.id, username=user.username, email=user.email, first_name=user.first_name
    )


def resp_to_users(resp):
    data = json.loads(resp.content.decode("utf-8"))
    pks = [d["id"] for d in data]
    users = []
    for pk in pks:
        # get+append to preserve order of resp'd data
        users.append(User.objects.get(pk=pk))
    return users


def my_view(request):
    names = "nomatch", "Bret", "Jemaine", "Murray"
    filters = CharFilter("username") | (
        EmailFilter("email") & ChoiceFilter("first_name", choices=zip(names, names))
    )
    try:
        query = filters.parse(request.GET)
        users = User.objects.filter(query)
        native = [user_to_native(u) for u in users]
        return HttpResponse(json.dumps(native))
    except InvalidData as ex:
        return HttpResponseBadRequest(json.dumps(ex.errors))


class FullStackTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.u1 = User.objects.create(
            username="one", email="one@example.org", first_name="Bret"
        )
        self.u2 = User.objects.create(
            username="two", email="two@example.org", first_name="Jemaine"
        )
        self.u3 = User.objects.create(
            username="three", email="three@example.org", first_name="Murray"
        )

    def test_request_no_params(self):
        """
        With no required filters, we present no filtering arguments and get all
        users returned.
        """
        request = self.factory.get("/my_view/")
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1, self.u2, self.u3]

        assert resp.status_code == 200
        assert actual == expected

    def test_request_irrelevant_params(self):
        """
        Filternaut doesn't care if the 'foo' param is present
        """
        request = self.factory.get("/my_view/?foo=bar")
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1, self.u2, self.u3]

        assert resp.status_code == 200
        assert actual == expected

    def test_request_nonmatching_param(self):
        """
        Filter on a non-existing username. Nothing comes back.
        """
        data = {"username": "nomatch"}
        request = self.factory.get("/my_view/", data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_but_nonmatching(self):
        """
        Provide both ANDed filters, but with values that don't match anything.
        Expect no results.
        """
        data = {"email": "nomatch@example.org", "first_name": "nomatch"}
        request = self.factory.get("/my_view/", data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_mixed_matching(self):
        """
        Provide both ANDed filters, with a matching email, but non-matching
        first name. Expect no results.
        """
        data = {
            "email": "one@example.org",  # a match
            "first_name": "nomatch",
        }  # a non-match
        request = self.factory.get("/my_view/", data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_matching(self):
        """
        Provide both ANDed filters. Both are matching. Expect one matching
        result.
        """
        data = {"email": "one@example.org", "first_name": "Bret"}
        request = self.factory.get("/my_view/", data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1]

        assert resp.status_code == 200
        assert actual == expected

    def test_invalid_filters(self):
        """
        Get some validation errors.
        """
        data = {"email": "not an email", "first_name": "not a choice"}
        request = self.factory.get("/my_view/", data=data)
        resp = my_view(request)
        errors = json.loads(resp.content.decode("utf-8"))

        assert resp.status_code == 400
        assert "email" in errors
        assert "first_name" in errors



from django.forms import (
    BooleanField,
    CharField,
    ChoiceField,
    ComboField,
    DateField,
    DateTimeField,
    DecimalField,
    EmailField,
    FilePathField,
    FloatField,
    ImageField,
    IntegerField,
    MultipleChoiceField,
    MultiValueField,
    NullBooleanField,
    RegexField,
    SlugField,
    SplitDateTimeField,
    TimeField,
    TypedChoiceField,
    URLField,
    ValidationError,
)

from filternaut import Filter, InvalidData
from filternaut.util import is_listlike, ordinal_suffix

# note IPAddressField, GenericIPAddressField and TypedMultipleChoiceField are
# conditionally imported later in this file; they are not available in all
# versions of Django.


__all__ = [
    "BooleanFilter",
    "CharFilter",
    "ChoiceFilter",
    "ComboFilter",
    "DateFilter",
    "DateTimeFilter",
    "DecimalFilter",
    "EmailFilter",
    "FilePathFilter",
    "FloatFilter",
    "IPAddressFilter",
    "ImageFilter",
    "FieldFilter",
    "IntegerFilter",
    "MultiValueFilter",
    "MultipleChoiceFilter",
    "NullBooleanFilter",
    "RegexFilter",
    "SlugFilter",
    "SplitDateTimeFilter",
    "TimeFilter",
    "TypedChoiceFilter",
    "URLFilter",
]


class FieldFilter(Filter):
    """
    FieldFilters use a django.forms.Field to clean their input value when
    generating a Q object.

    This class is designed to be extended by subclasses which provide their own
    form-field instances. However, you could use it in combination with a
    custom field like so::

        filter = FieldFilter(SpecialField(), dest='...')
    """

    def __init__(self, dest, field, **kwargs):
        self.field = field
        super().__init__(dest, **kwargs)

    def clean(self, value):
        if not is_listlike(value):
            return self.field.clean(value)

        cleaned = []
        for n, v in enumerate(value):
            try:
                cleaned.append(self.field.clean(v))
            except ValidationError as ex:
                suffix = ordinal_suffix(n + 1)
                raise ValidationError(
                    f"Validation error at {n + 1}{suffix} item: {ex.message}"
                ) from ex

        # use the type of `value` so that arguing a tuple returns a
        # tuple, list→list, etc.
        return type(value)(cleaned)


# -- mixtures of fieldfilter and django fields requiring additional arguments


class ChoiceFilter(FieldFilter):
    def __init__(self, dest, choices, *args, **kwargs):
        field = ChoiceField(choices=choices)
        super().__init__(dest, field=field, *args, **kwargs)


class RegexFilter(FieldFilter):
    def __init__(self, dest, regex, *args, **kwargs):
        field = RegexField(regex=regex)
        super().__init__(dest, field=field, *args, **kwargs)


class FilePathFilter(FieldFilter):
    def __init__(self, dest, path, *args, **kwargs):
        field = FilePathField(path=path)
        super().__init__(dest, field=field, *args, **kwargs)


class BooleanFilter(FieldFilter):
    """
    BooleanField required=True does not have sanely.
    """

    def __init__(self, dest, **kwargs):
        field = BooleanField(required=False)
        super().__init__(dest, field=field, **kwargs)


class ComboFilter(FieldFilter):
    def __init__(self, dest, fields, **kwargs):
        field = ComboField(fields=fields)
        super().__init__(dest, field=field, **kwargs)


# -- simple mixtures of fieldfilter and django fields.


class FieldMixin(object):
    def __init__(self, dest, **kwargs):
        # none-to-isnull converts None-values into __isnull=True. We've got to
        # set required=False on the field to get a None out of it.
        none_to_isnull = kwargs.get("none_to_isnull", False)
        field_required = none_to_isnull is False
        field = self.field_class(required=field_required)

        super().__init__(dest, field=field, **kwargs)


class CharFilter(FieldMixin, FieldFilter):
    field_class = CharField


class DateFilter(FieldMixin, FieldFilter):
    field_class = DateField


class DateTimeFilter(FieldMixin, FieldFilter):
    field_class = DateTimeField


class DecimalFilter(FieldMixin, FieldFilter):
    field_class = DecimalField


class EmailFilter(FieldMixin, FieldFilter):
    field_class = EmailField


class FloatFilter(FieldMixin, FieldFilter):
    field_class = FloatField


class ImageFilter(FieldMixin, FieldFilter):
    field_class = ImageField


class IntegerFilter(FieldMixin, FieldFilter):
    field_class = IntegerField


class MultiValueFilter(FieldMixin, FieldFilter):
    field_class = MultiValueField


class MultipleChoiceFilter(FieldMixin, FieldFilter):
    field_class = MultipleChoiceField


class NullBooleanFilter(FieldMixin, FieldFilter):
    field_class = NullBooleanField


class SlugFilter(FieldMixin, FieldFilter):
    field_class = SlugField


class SplitDateTimeFilter(FieldMixin, FieldFilter):
    field_class = SplitDateTimeField


class TimeFilter(FieldMixin, FieldFilter):
    field_class = TimeField


class TypedChoiceFilter(FieldMixin, FieldFilter):
    field_class = TypedChoiceField


class URLFilter(FieldMixin, FieldFilter):
    field_class = URLField


# -- mixtures of fieldfilter and django fields which were not present in older
# -- django versions

try:
    # django 1.3 and older did not have GenericIPAddressField or
    # TypedMultipleChoiceField, so we must conditionally define our matching
    # Filters.
    from django.forms import GenericIPAddressField, TypedMultipleChoiceField
except ImportError:
    pass
else:

    class GenericIPAddressFilter(FieldMixin, FieldFilter):
        field_class = GenericIPAddressField

    class TypedMultipleChoiceFilter(FieldMixin, FieldFilter):
        field_class = TypedMultipleChoiceField

    __all__.extend(("GenericIPAddressFilter", "TypedMultipleChoiceFilter"))

try:
    # django 1.9 and later drop support for IPAddressField
    from django.forms import IPAddressField
except ImportError:
    pass
else:

    class IPAddressFilter(FieldMixin, FieldFilter):
        field_class = IPAddressField

    __all__.extend(("IPAddressField",))



from collections import defaultdict, deque

from filternaut.util import is_listlike, ordinal_suffix


def test_ordinal_suffixes():
    for base in (0, 10, 20, 50, 100, 200, 500):
        assert ordinal_suffix(base + 0) == "th"
        assert ordinal_suffix(base + 1) == "st"
        assert ordinal_suffix(base + 2) == "nd"
        assert ordinal_suffix(base + 3) == "rd"
        for n in (4, 5, 6, 7, 8, 9):
            assert ordinal_suffix(base + n) == "th"
        assert ordinal_suffix(base + 10) == "th"


def test_is_listlike():
    assert is_listlike([1, 2, 3]) is True
    assert is_listlike((1, 2, 3)) is True
    assert is_listlike({1, 2, 3}) is True
    assert is_listlike(deque((1, 2, 3))) is True
    assert is_listlike(range(1, 4)) is True
    assert is_listlike((n for n in (1, 2, 3))) is True

    # a few intermediary cases for dicts. the dict itself isn't counted as
    # listlike, but its keys, values, or pairs thereof are.
    assert is_listlike({"foo": "bar"}.keys()) is True
    assert is_listlike({"foo": "bar"}.values()) is True
    assert is_listlike({"foo": "bar"}.items()) is True

    assert is_listlike(0) is False
    assert is_listlike(False) is False
    assert is_listlike({"foo": "bar"}) is False
    assert is_listlike(defaultdict(list, foo="bar")) is False
    assert is_listlike("text") is False
    assert is_listlike(b"bytes") is False



from unittest import TestCase

from django.http import QueryDict

from filternaut.filters import FieldFilter, Filter


class ExtensionTests(TestCase):
    """
    Check that folk can extend filternaut in useful ways.
    """

    def test_with_custom_field(self):
        """
        Custom fields should work with FieldFilter.
        """

        class DoublingField(object):
            def clean(self, value):
                return int(value) * 2

        filter = FieldFilter("integer", field=DoublingField())
        query = filter.parse({"integer": 3})

        actual = dict(query.children)
        expected = {"integer": 6}
        assert actual == expected

    def test_basic_filter_with_subclassed_clean(self):
        """
        Basic filters should let their source values be cleaned.
        """

        class PotatoFilter(Filter):
            def clean(self, value):
                return "potato"

        filter = PotatoFilter("fieldname")
        assert filter.clean("anything") == "potato"

    def test_specialcase_multiple_source_values(self):
        """
        If multiple source values are presented to parse, make sure we allow
        handling the picking of a single value.
        """

        class BiggestValueFilter(Filter):
            def get_source_value(self, key, data, many=False):
                values = data.getlist(key)
                return max(map(int, values))

        data = QueryDict("field=3&field=6&field=1&field=2")
        filter = BiggestValueFilter("field")
        query = filter.parse(data)
        actual = dict(query.children)
        expected = {"field": 6}
        assert actual == expected



from collections.abc import Iterable, Mapping

_suffixes = ("th", "st", "nd", "rd") + ("th",) * 6


def ordinal_suffix(n):
    """
    1→st, 2→nd, 3→rd, 4→th...
    """
    return _suffixes[n % 10]


def is_listlike(val):
    """
    True if `val` is "like a list".

    This is an arbitrary call. Included are 1-wide data structures you can
    iterate through (list, tuple, deque, generator), but not dicts because you
    only iterate through the keys (though the output of dict.items() *is*
    "listlike"), and not bytes or strings, because while iterating each
    character is handy, you never want to do that in the same context as
    iterating lists. Or that's my claim anyway.
    """
    return isinstance(val, Iterable) and not isinstance(val, (str, bytes, Mapping))



from textwrap import dedent

from setuptools import setup

setup(
    name="django-filternaut",
    version="1.0.0",
    author="Ben Graham",
    author_email="bgraham@bgraham.com.au",
    description=dedent("""\
        Construct arbitrarily complex Django "Q" filters from flat data, such
        as query parameters.
    """),
    long_description=open("README.rst").read(),
    url="https://github.com/bennullgraham/filternaut",
    license="BSD",
    packages=["filternaut"],
    install_requires=["six>=1.9.0", "django>=4.2"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Utilities",
        "Framework :: Django",
    ],
)



"""
This module ends up looking similar to django.utils.tree. It has a simpler
design allowed by the simpler use case. In particular there are
only ever two children, `left' and `right', which in turn allows the 'operator'
(what django terms a 'connector') to be literally a function from stdlib's
operator.

Finally, there are separate leaf and tree classes. This increases sanity when
these classes are extended to become filters.
"""

from copy import deepcopy
from operator import and_, or_


class Tree(object):
    # subclasses could specify a tree-subclass here. For example,
    # MyLeaf.tree_class = MyTree. A value of None will default to the current
    # class at the time of construction.
    tree_class = None

    def __init__(self, operator, left, right):
        assert operator in (or_, and_)
        self.operator = operator
        self.left = left
        self.right = right

    def __iter__(self):
        yield from self.left
        yield from self.right

    def __and__(self, other):
        klass = getattr(self, "tree_class") or self.__class__
        return klass(operator=and_, left=self, right=other)

    def __or__(self, other):
        klass = getattr(self, "tree_class") or self.__class__
        return klass(operator=or_, left=self, right=other)

    def copy(self):
        return deepcopy(self)


class Leaf(object):
    # subclasses could specify a tree-subclass here. For example,
    # MyLeaf.tree_class = MyTree.
    tree_class = Tree

    def __iter__(self):
        yield self

    def __and__(self, other):
        return self.tree_class(operator=and_, left=self, right=other)

    def __or__(self, other):
        return self.tree_class(operator=or_, left=self, right=other)

    def copy(self):
        return deepcopy(self)
